# https://github.com/tomdee/president

from copy import copy
from operator import attrgetter
import random
import sys

from blessings import Terminal

from jass_bot.mcts import GameState, Card, ismcts, Deck

# TODO:
# Make it faster? https://cardsource.readthedocs.org/en/latest/
# Make it look better - https://github.com/worldveil/deuces
# Let it play int he real world - http://arnab.org/blog/so-i-suck-24-automating-card-games-using-opencv-and-python
#                               - https://rdmilligan.wordpress.com/2014/08/30/playing-card-detection-using-opencv-mark-v/
# Allow it to work with any number of players (pass handling become more complicated)
# WOrk with Jokers

# run exclusivity
# Make sure sorted
CLEAN_PACK = Deck().cards
term = Terminal()


class PresidentGameState(GameState):
    """
    A state of the game President.
    See https://en.wikipedia.org/wiki/President_(card_game)
    """

    def __init__(self):
        """
        Initialise the game state.
        Always have 2 players

        """
        GameState.__init__(self)
        self.player_to_move = 0
        self.player_hands = [[], []]
        self.discards = []
        self.on_the_table = []
        self.combo_size = 0
        self.consecutive_mode = 0
        self.straight_length = 0

    def clone(self):
        """ Create a deep clone of this game state.
        """
        st = PresidentGameState()
        st.player_to_move = self.player_to_move
        st.player_hands = [copy(self.player_hands[0]), copy(self.player_hands[1])]
        st.discards = copy(self.discards)
        st.on_the_table = copy(self.on_the_table)
        st.combo_size = self.combo_size
        st.consecutive_mode = self.consecutive_mode
        st.straight_length = self.straight_length

        return st

    def clone_and_randomize(self, observer):
        """ Create a deep clone of this game state, randomizing any information
        not visible to the specified observer player.
        """
        st = self.clone()

        # The observer can see his own hand and the cards in the current trick,
        # and can remember the cards played in previous tricks
        flattened_current_trick = [a for b in st.on_the_table for a in b]
        seen_cards = set(st.player_hands[observer] + st.discards + flattened_current_trick)

        # The observer can't see the rest of the deck
        unseen_cards = [card for card in CLEAN_PACK
                        if card not in seen_cards]

        # deal the unseen cards to the player
        random.shuffle(unseen_cards)

        for p in (0, 1):
            if p != observer:
                # deal cards to player p. The players start with 17 cards, so
                # with two players there are 34 cards in total.
                # The player should be dealt 34 - discards - num cards in other players hand.

                # Store the size of player p's hand
                num_cards = 34 - len(st.discards) - len(st.player_hands[observer])

                # Give player p the first num_cards unseen cards
                st.player_hands[p] = sorted(unseen_cards[:num_cards],
                                            key=attrgetter('rank', 'suit'))

        return st

    def get_next_player(self, p):
        return (p + 1) % self.number_of_players

    def _deal(self):
        """
        Deal the cards
        """
        # Construct a deck, shuffle it, and deal it to the players
        deck = Deck()
        deck.shuffle()

        # Player zero gets the first 17 cards
        self.player_hands[0] = sorted(deck.cards[:17],
                                      key=attrgetter('rank', 'suit'))

        # Player one gets the last 17 cards
        self.player_hands[1] = sorted(deck.cards[-17:],
                                      key=attrgetter('rank', 'suit'))

    def do_move(self, move):
        """ update a state by carrying out the given move.
            Must update player_to_move.
        """
        # If the move is "PASS" then the current trick is over
        if move is "PASS":
            # Trick over so update the game state
            self.discards.extend([a for b in self.on_the_table for a in b])
            self.on_the_table = []
            self.straight_length = 0
            self.consecutive_mode = 0
            self.combo_size = 0
            self.player_to_move = self.get_next_player(self.player_to_move)
        else:
            if len(move) > 1:
                # The player put down multiple cards - add modes
                if move[0].rank == move[1].rank:
                    self.combo_size = len(move)
                else:
                    self.straight_length = len(move)

            # On the second move, check for CONSECUTIVE mode
            if len(self.on_the_table) == 1:
                # Check the highest card (TODO assume they are ordered) on the first hand.
                if self.on_the_table[0][-1].rank + 1 == move[0].rank:
                    self.consecutive_mode = 1

            # Store the played card in the current trick
            self.on_the_table.append(move)

            # Remove the card from the player's hand
            # TODO - this is probably slow
            for card in move:
                self.player_hands[self.player_to_move].remove(card)

            if self.player_hands[self.player_to_move]:
                # Only change players if the current player didn't just finish
                # If the current player finished, then we want to generate no
                # moves for that player to signal the end of hte game
                self.player_to_move = self.get_next_player(self.player_to_move)

    def get_moves(self):
        """
        Get all possible moves from this state.
        """

        hand = sorted(self.player_hands[self.player_to_move],
                      key=attrgetter('rank', 'suit'))
        if not hand:
            # If there are no moves left, then return the empty list.
            return hand
        else:
            if not self.on_the_table:
                # May lead a trick with any card. Can't pass - that would be silly.
                # Moves may involve multiple cards, so return a list of lists.
                candidate_cards = hand
            else:
                # Start by picking out just the higher cards. Card rank needs to be strictly greater.
                # Grab the rank of the last card from the last trick.
                minimum_rank = self.on_the_table[-1][-1].rank + 1

                if self.consecutive_mode:
                    if self.straight_length > 0:
                        # This is a straight - therefore we have a min and max range
                        candidate_cards = [card for card in hand if
                                           card.rank >= minimum_rank and card.rank < minimum_rank + self.straight_length]
                    else:
                        # Not a straight - therefore an exact rank is required
                        candidate_cards = [card for card in hand if card.rank == minimum_rank]
                else:
                    candidate_cards = [card for card in hand if card.rank >= minimum_rank]

            moves = []
            # TODO - alternative implementation for finding straights -
            # Split the deck by suit, and only both search suits that have more than three cards.

            for index, card in enumerate(candidate_cards):
                # Make a single loop through the candidate cards and include
                # not just the valid single card plays but also the DUBS, TRIPS, QUADS and RUNS.
                straight_found = [card]

                if self.combo_size == 0 and self.straight_length == 0:
                    # Always include the single cards when not in combo mode
                    moves.append([card])

                if not self.on_the_table or self.combo_size > 0 or self.straight_length > 0:
                    # It's either the first hand, or playing to a straight or combo. There's more work to do...
                    # Now get the list of the higher cards
                    next_cards = candidate_cards[index + 1:]

                    for next_cards_index, next_card in enumerate(next_cards):

                        if not self.on_the_table or (self.combo_size > 0 and self.straight_length == 0):
                            # Start looking ahead to find combos
                            if card.rank == next_card.rank:
                                if not self.on_the_table or self.combo_size == 2:
                                    # A pair
                                    moves.append([card, next_card])

                                if not self.on_the_table or self.combo_size == 3:
                                    # This is the second next_card - must be TRIPS
                                    if next_cards_index == 1:
                                        moves.append([card, next_cards[0], next_cards[1]])

                                if not self.on_the_table or self.combo_size == 4:
                                    # This is the third next_card - must be QUADS
                                    if next_cards_index == 2:
                                        moves.append([card, next_cards[0], next_cards[1], next_cards[2]])

                        if not self.on_the_table or self.straight_length > 0:
                            # Start looking ahead to find straights
                            # Check the last card found in the run found so far. Add the card to the list if it's good.
                            if straight_found[-1].rank + 1 == next_card.rank and straight_found[
                                -1].suit == next_card.suit:
                                straight_found.append(next_card)
                                if not self.on_the_table and len(straight_found) >= 3:
                                    # Opening move and the straight is at least three long. Record it.
                                    moves.append(copy(straight_found))
                                elif self.on_the_table and len(straight_found) == self.straight_length:
                                    # A straight long enough to play on a previous straight has been found.
                                    moves.append(straight_found)

            # Can always pass.
            return moves + ["PASS"]

    def get_result(self, player):
        """
        Get the game result from the viewpoint of player.
        """
        # If the play has nothing in their hand then they've won.
        return 1 if not self.player_hands[player] else 0

    def __repr__(self):
        """ Return a human-readable representation of the state
        """
        result = "P%i: %s Consec: %s Combo: %s Straight: %s Trick: %s Discards: %s" % (
            self.player_to_move,
            self.player_hands[self.player_to_move],
            # self.player_hands,
            self.consecutive_mode,
            self.combo_size,
            self.straight_length,
            self.on_the_table,
            self.discards)
        return result


def play_self():
    """ Play a sample game between two ismcts players.
    """
    state = PresidentGameState()
    state._deal()

    while state.get_moves():
        print
        str(state)
        # Use different numbers of iterations (simulations, tree nodes) for different players
        if state.player_to_move == 0:
            m = ismcts(rootstate=state, itermax=1000, verbose=False)
        else:
            m = ismcts(rootstate=state, itermax=100, verbose=False)
        print
        "Best Move: " + str(m) + "\n"
        state.do_move(m)

    for p in (0, 1):
        if state.get_result(p) > 0:
            print
            "Player " + str(p) + " wins!"


def get_card(prompt):
    card = None

    while not card:
        try:
            card = Card(raw_input(prompt).upper())
        except Exception:
            print
            "Bad card - try again"
    return card


def get_int(prompt):
    input = None

    while input is None:
        try:
            input = int(raw_input(prompt))
        except Exception:
            print
            "Bad input - try again"
    return input


def play_game():
    """ Play a game between one human and one AI
    """
    state = PresidentGameState()
    # state.player_hands[0] = [
    #                          Card('3H'),
    #                          Card('4H'),
    #                          Card('5C'),
    #                          Card('5D'),
    #                          Card('6H'),
    #                          Card('7H'),
    #                          Card('8C'),
    #                          Card('9H'),
    #                          Card('JH'),
    #                          Card('QH'),
    #                          Card('KC'),
    #                          Card('AH'),
    #                          Card('2H'),
    #                          Card('QD'),
    #                          Card('KS'),
    #                          Card('AD'),
    #                          Card('2D')]
    state.player_hands[0] = [
        Card('9d'),
        Card('js'),
        Card('qc'),
        Card('3d'),
        Card('as'),
        Card('2s'),
        Card('ts'),
        Card('jd'),
        Card('6s'),
        Card('9s'),
        Card('5h'),
        Card('7d'),
        Card('8s'),
        Card('4d'),
        Card('qs'),
        Card('tc'),
        Card('7s')]

    TOTAL_CARDS = 17
    while len(state.player_hands[0]) < TOTAL_CARDS:
        card = get_card("Enter card %s/%s: " % (len(state.player_hands[0]) + 1, TOTAL_CARDS))
        if card not in state.player_hands[0]:
            state.player_hands[0].append(card)
        else:
            print
            "Duplicate card - ignoring"

    print
    "All done"
    print
    state.player_hands[0]

    while True:
        print
        str(state)
        # Use different numbers of iterations (simulations, tree nodes) for different players
        if state.player_to_move == 0:
            m = ismcts(rootstate=state, itermax=10000, verbose=False)
            print
            "Best Move: " + str(m) + "\n"
            state.do_move(m)
            if not state.player_hands[0]:
                # No cards left - the end
                break

        else:
            confirmed_move = False
            move = []
            while not confirmed_move:
                num_cards = get_int("How many cards did the other player play?")
                move = []
                if num_cards == 0:
                    move = "PASS"
                else:
                    for i in range(num_cards):
                        card = get_card("What was the card?")
                        move.append(card)

                user_input = raw_input("Did the player play this (y/n)? %s" % move)
                if user_input is "y" or user_input is "Y":
                    confirmed_move = True
                else:
                    print
                    "Let's try again..."

            if move is not "PASS":
                state.player_hands[1].extend(move)
            state.do_move(move)

            if move is not "PASS":
                state.player_to_move = state.get_next_player(state.player_to_move)

    for p in (0, 1):
        if state.get_result(p) > 0:
            print
            "Player " + str(p) + " wins!"


def test_iters():
    # I want to know how many iteractions are good.
    # I'll play a player that does only 10 iterations, then keep pushing mine up to see how many wins/losses

    # Play this many games at each iter value
    NUM_GAMES = 100

    for iterations in (1, 10, 100, 1000, 10000):

        win_counts = [0, 0]
        for game_num in range(NUM_GAMES):
            state = PresidentGameState()
            state._deal()

            with term.location(0, term.height - 1):
                print("(%s) Game number %s/%s" % (iterations, game_num, NUM_GAMES)),
                sys.stdout.flush()

            while state.get_moves():
                # Use different numbers of iterations (simulations, tree nodes) for different players
                if state.player_to_move == 0:
                    m = ismcts(rootstate=state, itermax=iterations, verbose=False, quiet=True)
                else:
                    m = ismcts(rootstate=state, itermax=1, verbose=False, quiet=True)
                # print "Best Move: " + str(m) + "\n"
                state.do_move(m)

            # print "Results: P0: %s P1: %s" % (state.get_result(0), state.get_result(1))
            # if (state.get_result(0) and state.get_result(1)) or not (state.get_result(0) or state.get_result(1)):
            #     # Nobody won or both players won
            #     print state
            for p in (0, 1):
                win_counts[p] += state.get_result(p)
        term.clear_eol()
        print
        "Iteration %s - %s" % (iterations, win_counts)


if __name__ == "__main__":
    play_game()
    # play_self()
    # test_iters()



# https://github.com/tomdee/president

from math import *
from operator import attrgetter
import random
from blessings import Terminal
import sys
term = Terminal()

class GameState:
    """ A state of the game, i.e. the game board. These are the only functions which are
        absolutely necessary to implement ismcts in any imperfect information game,
        although they could be enhanced and made quicker, for example by using a
        GetRandomMove() function to generate a random move during rollout.
        By convention the players are numbered 1, 2, ..., self.number_of_players.
    """

    def __init__(self):
        self.number_of_players = 2
        self.player_to_move = 1

    def get_next_player(self, p):
        """ Return the player to the left of the specified player
        """
        return (p % self.number_of_players) + 1

    def clone(self):
        """ Create a deep clone of this game state.
        """
        st = GameState()
        st.player_to_move = self.player_to_move
        return st

    def clone_and_randomize(self, observer):
        """ Create a deep clone of this game state, randomizing any information not visible to the specified observer player.
        """
        return self.clone()

    def do_move(self, move):
        """ update a state by carrying out the given move.
            Must update player_to_move.
        """
        self.player_to_move = self.get_next_player(self.player_to_move)

    def get_moves(self):
        """ Get all possible moves from this state.
        """
        pass

    def get_result(self, player):
        """ Get the game result from the viewpoint of player.
        """
        pass

    def __repr__(self):
        """ Don't need this - but good style.
        """
        pass

class Deck:
    """
    A deck of cards
    """
    def __init__(self):
        self.cards = [Card(rank, suit) for rank in xrange(3, 15 + 1) for suit in
                ['C', 'D', 'H', 'S']]

    def shuffle(self):
        random.shuffle(self.cards)


NAMES = "???3456789TJQKA2"
class Card(object):
    """
    A playing card, with rank and suit.
    rank must be an integer between 3 and 15 inclusive (Jack=11, Queen=12, King=13, Ace=14)
    suit must be a string of length 1, one of 'C' (Clubs), 'D' (Diamonds), 'H' (Hearts) or 'S' (Spades)
    """
    def __init__(self, rank=None, suit=None):
        if not suit:
            suit = rank[1].upper()
            rank = NAMES.index(rank[0].upper())
            if rank not in range(2, 15 + 1):
                raise Exception("Invalid rank")
            if suit not in ['C', 'D', 'H', 'S']:
                raise Exception("Invalid suit")

        self.rank = rank
        self.suit = suit

    def __repr__(self):
        return NAMES[self.rank] + PRETTY_SUITS[self.suit]

    def __eq__(self, other):
        return self.rank == other.rank and self.suit == other.suit

    def __ne__(self, other):
        return self.rank != other.rank or self.suit != other.suit


PRETTY_SUITS = {
        "S" : u"\u2660".encode('utf-8'), # spades
        "H" : u"\u2764".encode('utf-8'), # hearts
        "D" : u"\u2666".encode('utf-8'), # diamonds
        "C" : u"\u2663".encode('utf-8') # clubs
    }

class Node:
    """
    A node in the game tree. Note wins is always from the viewpoint of player_just_moved.
    """

    def __init__(self, move=None, parent=None, player_just_moved=None):
        # the move that got us to this node - "None" for the root node
        self.move = move
        # "None" for the root node
        self.parent_node = parent
        self.child_nodes = []
        self.wins = 0
        self.visits = 0
        self.avails = 1
        # the only part of the state that the Node needs later
        self.player_just_moved = player_just_moved

    def get_untried_moves(self, legal_moves):
        """
        Return the elements of legal_moves for which this node does not have children.
        """

        # Find all moves for which this node *does* have children
        tried_moves = [child.move for child in self.child_nodes]

        # Return all moves that are legal but have not been tried yet
        return [move for move in legal_moves if move not in tried_moves]

    def ucb_select_child(self, legal_moves, exploration=0.7):
        """
        Use the UCB1 formula to select a child node, filtered by the given list of legal moves.
        exploration is a constant balancing between exploitation and exploration, with default value 0.7 (approximately sqrt(2) / 2)
        """

        # Filter the list of children by the list of legal moves
        legal_children = [child for child in self.child_nodes if
                          child.move in legal_moves]

        # Get the child with the highest UCB score
        s = max(legal_children, key=lambda c: float(c.wins) / float(
            c.visits) + exploration * sqrt(log(c.avails) / float(c.visits)))

        # update availability counts -- it is easier to do this now than during backpropagation
        for child in legal_children:
            child.avails += 1

        # Return the child selected above
        return s

    def add_child(self, m, p):
        """
        Add a new child node for the move m.
        Return the added child node
        """
        n = Node(move=m, parent=self, player_just_moved=p)
        self.child_nodes.append(n)
        return n

    def update(self, terminal_state):
        """
        update this node - increment the visit count by one, and increase the win count by the result of terminal_state for self.player_just_moved.
        """
        self.visits += 1
        if self.player_just_moved is not None:
            self.wins += terminal_state.get_result(self.player_just_moved)

    def __repr__(self):
        return "[M:%s W/V/A: %4i/%4i/%4i]" % (
        self.move, self.wins, self.visits, self.avails)

    def tree_to_string(self, indent):
        """
        Represent the tree as a string, for debugging purposes.
        """
        s = self.indent_string(indent) + str(self)
        for c in sorted(self.child_nodes, key=attrgetter('visits', 'wins')):
            s += c.tree_to_string(indent + 1)
        return s

    @staticmethod
    def indent_string(indent):
        s = "\n"
        for i in range(1, indent + 1):
            s += "| "
        return s

    def children_to_string(self):
        s = ""
        for c in sorted(self.child_nodes, key=attrgetter('visits', 'wins')):
            s += str(c) + "\n"
        return s


def ismcts(rootstate, itermax, verbose=False, quiet=False):
    """
    Conduct an ismcts search for itermax iterations starting from rootstate.
    Return the best move from the rootstate.
    """

    rootnode = Node()
    if len(rootstate.get_moves()) > 1:
        # There are moves. Simulate them

        for i in range(itermax):
            node = rootnode

            # Determinize
            state = rootstate.clone_and_randomize(rootstate.player_to_move)

            # Select
            while state.get_moves() != [] and node.get_untried_moves(
                    state.get_moves()) == []:  # node is fully expanded and non-terminal
                node = node.ucb_select_child(state.get_moves())
                state.do_move(node.move)

            # Expand
            untried_moves = node.get_untried_moves(state.get_moves())
            if untried_moves:  # if we can expand (i.e. state/node is non-terminal)
                m = random.choice(untried_moves)
                player = state.player_to_move
                state.do_move(m)
                node = node.add_child(m, player)  # add child and descend tree

            # Simulate
            while state.get_moves():  # while state is non-terminal
                state.do_move(random.choice(state.get_moves()))

            # Backpropagate
            while node:  # backpropagate from the expanded node and work back to the root node
                node.update(state)
                node = node.parent_node

            if not quiet:
                with term.location(0, term.height - 1):
                    print("Iteration %s/%s - Best so far (%s)" % (i, itermax, max(rootnode.child_nodes, key=lambda
            c: c.visits).move)),
                    sys.stdout.flush()
    else:
        rootnode.add_child(rootstate.get_moves()[0], rootstate.player_to_move)

    # Output some information about the tree - can be omitted
    if verbose:
        term.clear_eol()
        print rootnode.tree_to_string(0)
    elif not quiet:
        term.clear_eol()
        print rootnode.children_to_string()

    return max(rootnode.child_nodes, key=lambda
        c: c.visits).move  # return the move that was most visited


#  try:
#         print x
#         time.sleep(.3)
#         x += 1
#     except KeyboardInterrupt: